@@ -1,4 +1,4 @@
-# Copyright 2011-2018, Damian Johnson and The Tor Project
+# Copyright 2011-2019, Damian Johnson and The Tor Project
 # See LICENSE for licensing information
 
 """
@@ -28,6 +28,9 @@ Library for working with the tor process.
     +- SocketError - Communication with the socket failed.
        +- SocketClosed - Socket has been shut down.
 
+  DownloadFailed - Inability to download a resource.
+    +- DownloadTimeout - Download timeout reached.
+
 .. data:: Runlevel (enum)
 
   Rating of importance used for event logging.
@@ -49,6 +52,10 @@ Library for working with the tor process.
   .. versionchanged:: 1.3.0
      Added the HEARTBEAT signal.
 
+  .. versionchanged:: 1.8.0
+     Added the ACTIVE and DORMANT signals. You can check for Tor support for
+     these signals with the **DORMANT_MODE** :data:`~stem.version.Requirement`
+
   ========================= ===========
   Signal                    Description
   ========================= ===========
@@ -60,6 +67,8 @@ Library for working with the tor process.
   **NEWNYM**                switch to new circuits, so new application requests don't share any circuits with old ones (this also clears our DNS cache)
   **CLEARDNSCACHE**         clears cached DNS results
   **HEARTBEAT**             trigger a heartbeat log message
+  **DORMANT**               enables *dormant mode*, during which tor will avoid cpu and network usage
+  **ACTIVE**                disables *dormant mode*
   ========================= ===========
 
 .. data:: Flag (enum)
@@ -72,6 +81,9 @@ Library for working with the tor process.
   .. versionchanged:: 1.5.0
      Added the NO_ED_CONSENSUS flag.
 
+  .. versionchanged:: 1.8.0
+     Added the STALE_DESC flag.
+
   =================== ===========
   Flag                Description
   =================== ===========
@@ -86,6 +98,7 @@ Library for working with the tor process.
   **NO_ED_CONSENSUS** relay's Ed25519 doesn't reflrect the consensus
   **RUNNING**         relay is currently usable
   **STABLE**          relay's suitable for long-lived circuits
+  **STALE_DESC**      relay descriptor is outdated and should be re-uploaded
   **UNNAMED**         relay isn't currently bound to a nickname
   **V2DIR**           relay supports the v2 directory protocol
   **VALID**           relay has been validated
@@ -458,16 +471,20 @@ Library for working with the tor process.
   .. versionchanged:: 1.6.0
      Added the QUERY_NO_HSDIR reason.
 
-  =================== ===========
-  HSDescReason        Description
-  =================== ===========
-  **BAD_DESC**        descriptor was unparseable
-  **QUERY_REJECTED**  hidden service directory refused to provide the descriptor
-  **UPLOAD_REJECTED** descriptor was rejected by the hidden service directory
-  **NOT_FOUND**       descriptor with the given identifier wasn't found
-  **QUERY_NO_HSDIR**  no hidden service directory was found
-  **UNEXPECTED**      failure type is unknown
-  =================== ===========
+  .. versionchanged:: 1.8.0
+     Added the QUERY_RATE_LIMITED reason.
+
+  ======================= ===========
+  HSDescReason            Description
+  ======================= ===========
+  **BAD_DESC**            descriptor was unparseable
+  **QUERY_REJECTED**      hidden service directory refused to provide the descriptor
+  **UPLOAD_REJECTED**     descriptor was rejected by the hidden service directory
+  **NOT_FOUND**           descriptor with the given identifier wasn't found
+  **QUERY_NO_HSDIR**      no hidden service directory was found
+  **QUERY_RATE_LIMITED**  request was throttled
+  **UNEXPECTED**          failure type is unknown
+  ======================= ===========
 
 .. data:: HSAuth (enum)
 
@@ -485,10 +502,12 @@ Library for working with the tor process.
   ================= ===========
 """
 
+import traceback
+
 import stem.util
 import stem.util.enum
 
-__version__ = '1.7.1'
+__version__ = '1.8.2'
 __author__ = 'Damian Johnson'
 __contact__ = 'atagar@torproject.org'
 __url__ = 'https://stem.torproject.org/'
@@ -518,6 +537,8 @@ __all__ = [
   'InvalidArguments',
   'SocketError',
   'SocketClosed',
+  'DownloadFailed',
+  'DownloadTimeout',
   'Runlevel',
   'Signal',
   'Flag',
@@ -698,6 +719,66 @@ class SocketClosed(SocketError):
   'Control socket was closed before completing the message.'
 
 
+class DownloadFailed(IOError):
+  """
+  Inability to download a resource. Python's urllib module raises
+  a wide variety of undocumented exceptions (urllib2.URLError,
+  socket.timeout, and others).
+
+  This wraps lower level failures in a common exception type that
+  retains their exception and `stacktrace
+  <https://docs.python.org/3/library/traceback.html>`_.
+
+  .. versionadded:: 1.8.0
+
+  :var str url: url we failed to download from
+  :var Exception error: original urllib exception
+  :var traceback stacktrace: original stacktrace
+  :var str stacktrace_str: string representation of the stacktrace
+  """
+
+  def __init__(self, url, error, stacktrace, message = None):
+    if message is None:
+      # The string representation of exceptions can reside in several places.
+      # urllib.URLError use a 'reason' attribute that in turn may referrence
+      # low level structures such as socket.gaierror. Whereas most exceptions
+      # use a 'message' attribute.
+
+      reason = str(error)
+
+      all_str_repr = (
+        getattr(getattr(error, 'reason', None), 'strerror', None),
+        getattr(error, 'reason', None),
+        getattr(error, 'message', None),
+      )
+
+      for str_repr in all_str_repr:
+        if str_repr and isinstance(str_repr, str):
+          reason = str_repr
+          break
+
+      message = 'Failed to download from %s (%s): %s' % (url, type(error).__name__, reason)
+
+    super(DownloadFailed, self).__init__(message)
+
+    self.url = url
+    self.error = error
+    self.stacktrace = stacktrace
+    self.stacktrace_str = ''.join(traceback.format_tb(stacktrace))
+
+
+class DownloadTimeout(DownloadFailed):
+  """
+  Timeout reached while downloading this resource.
+
+  .. versionadded:: 1.8.0
+  """
+
+  def __init__(self, url, error, stacktrace, timeout):
+    message = 'Failed to download from %s: %0.1f second timeout reached' % (url, timeout)
+    super(DownloadTimeout, self).__init__(url, error, stacktrace, message)
+
+
 Runlevel = stem.util.enum.UppercaseEnum(
   'DEBUG',
   'INFO',
@@ -718,6 +799,7 @@ Flag = stem.util.enum.Enum(
   ('NO_ED_CONSENSUS', 'NoEdConsensus'),
   ('RUNNING', 'Running'),
   ('STABLE', 'Stable'),
+  ('STALE_DESC', 'StaleDesc'),
   ('UNNAMED', 'Unnamed'),
   ('V2DIR', 'V2Dir'),
   ('V3DIR', 'V3Dir'),
@@ -738,6 +820,8 @@ Signal = stem.util.enum.UppercaseEnum(
   'NEWNYM',
   'CLEARDNSCACHE',
   'HEARTBEAT',
+  'ACTIVE',
+  'DORMANT',
 )
 
 CircStatus = stem.util.enum.UppercaseEnum(
